import datetime
import logging
import os
import subprocess
from typing import Union
from urllib.parse import urlparse

import bbconf
import pephubclient
import peppy
import pypiper
import yaml
from bbconf.bbagent import BedBaseAgent
from bbconf.const import DEFAULT_LICENSE
from bbconf.models.base_models import FileModel
from eido import validate_project
from geniml.bbclient import BBClient
from pephubclient.helpers import MessageHandler as m
from pephubclient.helpers import is_registry_path

from bedboss._version import __version__
from bedboss.bedbuncher import run_bedbuncher
from bedboss.bedmaker.bedmaker import make_all
from bedboss.bedstat.bedstat import bedstat
from bedboss.const import BEDBOSS_PEP_SCHEMA_PATH, PKG_NAME, MAX_FILE_SIZE_QC
from bedboss.exceptions import BedBossException, QualityException
from bedboss.models import (
    BedClassificationUpload,
    BedSetAnnotations,
    FilesUpload,
    PlotsUpload,
    StatsUpload,
)
from bedboss.refgenome_validator.main import ReferenceValidator
from bedboss.refgenome_validator.utils import predict_from_compatibility_resutlts
from bedboss.skipper import Skipper
from bedboss.utils import (
    calculate_time,
    get_genome_digest,
    standardize_genome_name,
    run_initial_qc,
)
from bedboss.utils import standardize_pep as pep_standardizer
from bedboss.bedstat.r_service import RServiceManager

_LOGGER = logging.getLogger(PKG_NAME)


def requirements_check() -> None:
    """
    Check if all requirements are installed

    :return: None
    """
    _LOGGER.info("Checking requirements...")
    subprocess.run(
        [
            "bash",
            f"{os.path.dirname(os.path.abspath(__file__))}/scripts/requirements_test.sh",
        ]
    )


@calculate_time
def run_all(
    input_file: str,
    input_type: str,
    outfolder: str,
    genome: str,
    bedbase_config: Union[str, bbconf.BedBaseAgent],
    name: str = None,
    license_id: str = DEFAULT_LICENSE,
    rfg_config: str = None,
    narrowpeak: bool = False,
    check_qc: bool = True,
    validate_reference: bool = True,
    chrom_sizes: str = None,
    open_signal_matrix: str = None,
    ensdb: str = None,
    other_metadata: dict = None,
    just_db_commit: bool = False,
    force_overwrite: bool = False,
    update: bool = False,
    upload_qdrant: bool = False,
    upload_s3: bool = False,
    upload_pephub: bool = False,
    lite: bool = False,
    # Universes
    universe: bool = False,
    universe_method: str = None,
    universe_bedset: str = None,
    pm: pypiper.PipelineManager = None,
    r_service: RServiceManager = None,
    reference_genome_validator: ReferenceValidator = None,
) -> str:
    """
    Run bedboss: bedmaker -> bedqc -> bedclassifier -> bedstat -> upload to s3, qdrant, pephub, and bedbase.

    :param str input_file: Input file [required]
    :param str input_type: Input type [required] options: (bigwig|bedgraph|bed|bigbed|wig)
    :param str outfolder: Folder, where output should be saved  [required]
    :param str genome: genome_assembly of the sample. [required] options: (hg19, hg38, mm10) # TODO: add more
    :param str name: name of the sample (human-readable name, e.g. "H3K27ac in liver") [optional]
    :param Union[str, bbconf.BedBaseConf] bedbase_config: The path to the bedbase configuration file, or bbconf object.
    :param str license_id: license identifier [optional] (default: "DUO:0000042").; Find All licenses in bedbase.org
    :param str rfg_config: file path to the genome config file [optional]
    :param bool narrowpeak: whether the regions are narrow. Used to create bed file from bedgraph or bigwig
        (transcription factor implies narrow, histone mark implies broad peaks) [optional]
    :param bool check_qc: set True to run quality control during badmaking [optional] (default: True)
    :param bool validate_reference: set True to run genome reference validator
    :param str chrom_sizes: a full path to the chrom.sizes required for the bedtobigbed conversion [optional]
    :param str open_signal_matrix: a full path to the openSignalMatrix required for the tissue [optional]
    :param dict other_metadata: a dict containing all attributes from the sample
    :param str ensdb: a full path to the ensdb gtf file required for genomes not in GDdata [optional]
        (basically genomes that's not in GDdata)
    :param bool just_db_commit: whether just to commit the JSON to the database [Default: False]
    :param bool force_overwrite: force overwrite analysis [Default: False]
    :param bool update: whether to update the record in the database [Default: False] (if True, overwrites 'force_overwrite' and ignores it)
    :param bool upload_qdrant: whether to skip qdrant indexing [Default: False]
    :param bool upload_s3: whether to upload to s3
    :param bool upload_pephub: whether to push bedfiles and metadata to pephub [Default: False]
    :param bool lite: whether to run lite version of the pipeline [Default: False]

    :param bool universe: whether to add the sample as the universe [Default: False]
    :param str universe_method: method used to create the universe [Default: None]
    :param str universe_bedset: bedset identifier for the universe [Default: None]
    :param pypiper.PipelineManager pm: pypiper object
    :param RServiceManager r_service: RServiceManager object that will run R services
    :param reference_genome_validator: ReferenceValidator object that will validate reference genome compatibility
    :return str bed_digest: bed digest
    """
    if isinstance(bedbase_config, str):
        bbagent = BedBaseAgent(config=bedbase_config, init_ml=not lite)
    elif isinstance(bedbase_config, bbconf.BedBaseAgent):
        bbagent = bedbase_config
    else:
        raise BedBossException("Incorrect bedbase_config type. Exiting...")

    genome = standardize_genome_name(genome)

    _LOGGER.info(f"Input file = '{input_file}'")
    _LOGGER.info(f"Output folder = '{outfolder}'")
    _LOGGER.info(f"Sample genome = '{genome}'")

    if not pm:
        pm_out_folder = os.path.join(os.path.abspath(outfolder), "pipeline_manager")
        _LOGGER.info(f"Pipeline info folder = '{pm_out_folder}'")
        pm = pypiper.PipelineManager(
            name="bedboss-pipeline",
            outfolder=pm_out_folder,
            version=__version__,
            recover=True,
        )
        stop_pipeline = True
    else:
        stop_pipeline = False

    if urlparse(input_file).scheme in ("http", "https", "ftp"):
        _LOGGER.info(
            "Remote input detected. Running initial_qc to inspect remote file size."
        )
        file_size = run_initial_qc(input_file)

        if file_size > MAX_FILE_SIZE_QC:
            raise QualityException(
                f"Remote file size {file_size} exceeds the maximum allowed size of {bbconf.MAX_REMOTE_FILE_SIZE}. Please download the file and provide a local path."
            )
        else:
            _LOGGER.info(
                f"Remote file size {file_size} is within the allowed limit. Proceeding with processing."
            )

    bed_metadata = make_all(
        input_file=input_file,
        input_type=input_type,
        output_path=outfolder,
        genome=genome,
        rfg_config=rfg_config,
        narrowpeak=narrowpeak,
        check_qc=check_qc,
        chrom_sizes=chrom_sizes,
        lite=lite,
        pm=pm,
    )
    if not other_metadata:
        other_metadata = {"sample_name": name}

    if not update:
        other_metadata["original_file_name"] = os.path.basename(input_file)

    if lite:
        statistics_dict = {}
        statistics_dict["number_of_regions"] = len(bed_metadata.bed_object)
    else:
        statistics_dict = bedstat(
            bedfile=bed_metadata.bed_file,
            outfolder=outfolder,
            genome=genome,
            ensdb=ensdb,
            bed_digest=bed_metadata.bed_digest,
            open_signal_matrix=open_signal_matrix,
            just_db_commit=just_db_commit,
            rfg_config=rfg_config,
            pm=pm,
            r_service=r_service,
        )

    if "mean_region_width" not in statistics_dict:
        statistics_dict["mean_region_width"] = (
            bed_metadata.bed_object.mean_region_width()
        )

    statistics_dict["bed_compliance"] = bed_metadata.bed_compliance
    statistics_dict["data_format"] = bed_metadata.data_format.value

    if bed_metadata.bigbed_file:
        genome_digest = get_genome_digest(genome)
    else:
        genome_digest = None

    stats = StatsUpload(**statistics_dict)
    plots = PlotsUpload(**statistics_dict)

    if bed_metadata.bigbed_file:
        big_bed = FileModel(
            name="bigbedfile",
            title="BigBed file",
            path=bed_metadata.bigbed_file,
            description="Path to the bigbed file",
            thumbnail_path=None,
            file_digest=None,
        )
    else:
        big_bed = None
    files = FilesUpload(
        bed_file=FileModel(
            name="bedfile",
            title="BED file",
            path=bed_metadata.bed_file,
            description="Path to the BED file",
            thumbnail_path=None,
            file_digest=bed_metadata.bed_object.file_digest,
        ),
        bigbed_file=big_bed,
    )

    classification = BedClassificationUpload(
        name=name or bed_metadata.bed_digest,
        genome_digest=genome_digest,
        genome_alias=genome,
        bed_compliance=bed_metadata.bed_compliance,
        data_format=bed_metadata.data_format.value,
        compliant_columns=bed_metadata.compliant_columns,
        non_compliant_columns=bed_metadata.non_compliant_columns,
        header=bed_metadata.bed_object.header,
    )
    if not reference_genome_validator:
        reference_genome_validator = ReferenceValidator()
    if validate_reference:
        _LOGGER.info("Validating reference genome")
        ref_valid_stats = reference_genome_validator.determine_compatibility(
            bedfile=bed_metadata.bed_object, concise=True
        )
        predicted_alias, predicted_digest = predict_from_compatibility_resutlts(
            ref_valid_stats
        )
        if predicted_alias and predicted_digest:
            _LOGGER.info(
                f"Predicted genome: {predicted_alias} (digest: {predicted_digest})"
            )
            classification.genome_alias = predicted_alias
            classification.genome_digest = predicted_digest
    else:
        ref_valid_stats = None

    if update:
        bbagent.bed.update(
            identifier=bed_metadata.bed_digest,
            stats=stats.model_dump(exclude_unset=True),
            metadata=other_metadata,
            plots=plots.model_dump(exclude_unset=True),
            files=files.model_dump(exclude_unset=True),
            classification=classification.model_dump(exclude_unset=True),
            ref_validation=ref_valid_stats,
            license_id=license_id,
            upload_qdrant=upload_qdrant and not lite,
            upload_pephub=upload_pephub,
            upload_s3=upload_s3,
            local_path=outfolder,
            overwrite=True,
            processed=not lite,
            nofail=True,
        )
    else:
        bbagent.bed.add(
            identifier=bed_metadata.bed_digest,
            stats=stats.model_dump(exclude_unset=True),
            metadata=other_metadata,
            plots=plots.model_dump(exclude_unset=True),
            files=files.model_dump(exclude_unset=True),
            classification=classification.model_dump(exclude_unset=True),
            ref_validation=ref_valid_stats,
            license_id=license_id,
            upload_qdrant=upload_qdrant and not lite,
            upload_pephub=upload_pephub,
            upload_s3=upload_s3,
            local_path=outfolder,
            overwrite=force_overwrite,
            processed=not lite,
            nofail=True,
        )

    if universe:
        bbagent.bed.add_universe(
            bedfile_id=bed_metadata.bed_digest,
            bedset_id=universe_bedset,
            construct_method=universe_method,
        )

    if stop_pipeline:
        pm.stop_pipeline()

    _LOGGER.info(f"All done! Bed digest: {bed_metadata.bed_digest}")
    return bed_metadata.bed_digest


@calculate_time
def insert_pep(
    bedbase_config: str,
    output_folder: str,
    pep: Union[str, peppy.Project],
    bedset_id: str = None,
    bedset_name: str = None,
    rfg_config: str = None,
    license_id: str = DEFAULT_LICENSE,
    create_bedset: bool = False,
    bedset_heavy: bool = False,
    check_qc: bool = True,
    ensdb: str = None,
    just_db_commit: bool = False,
    force_overwrite: bool = False,
    update: bool = False,
    upload_s3: bool = False,
    upload_pephub: bool = False,
    upload_qdrant: bool = False,
    no_fail: bool = False,
    standardize_pep: bool = False,
    lite: bool = False,
    rerun: bool = False,
    pm: pypiper.PipelineManager = None,
) -> None:
    """
    Run all bedboss pipelines for all samples in the pep file.
    bedmaker -> bedqc -> bedstat -> qdrant_indexing -> bedbuncher

    :param str bedbase_config: bedbase configuration file path
    :param str output_folder: output statistics folder
    :param Union[str, peppy.Project] pep: path to the pep file or pephub registry path
    :param str bedset_id: bedset identifier
    :param str bedset_name: bedset name
    :param str rfg_config: path to the genome config file (refgenie)
    :param str license_id: license identifier [optional] (default: "DUO:0000042").; Find All licenses in bedbase.org
        This license will be used for bedfiles where license is not provided in PEP file
    :param bool create_bedset: whether to create bedset
    :param bool bedset_heavy: whether to use heavy processing (add all columns to the database)
    :param bool upload_qdrant: whether to upload bedfiles to qdrant
    :param bool check_qc: whether to run quality control during badmaking
    :param str ensdb: a full path to the ensdb gtf file required for genomes not in GDdata
    :param bool just_db_commit: whether save only to the database (Without saving locally )
    :param bool force_overwrite: whether to overwrite the existing record
    :param bool update: whether to update the record in the database. This option will overwrite the force_overwrite option. [Default: False]
    :param bool upload_s3: whether to upload to s3
    :param bool upload_pephub: whether to push bedfiles and metadata to pephub (default: False)
    :param bool upload_qdrant: whether to execute qdrant indexing
    :param bool no_fail: whether to raise an error if bedset was not added to the database
    :param bool lite: whether to run lite version of the pipeline
    :param bool standardize_pep: whether to standardize the pep file before processing by using bedms. (default: False)
    :param bool rerun: whether to rerun processed samples
    :param pypiper.PipelineManager pm: pypiper object
    :return: None
    """

    failed_samples = []
    processed_ids = []
    if isinstance(pep, peppy.Project):
        pass
    elif isinstance(pep, str):
        if is_registry_path(pep):
            pep = pephubclient.PEPHubClient().load_project(pep)
        else:
            pep = peppy.Project(pep)
    else:
        raise BedBossException("Incorrect pep type. Exiting...")

    if standardize_pep:
        pep = pep_standardizer(pep)

    if not pm:
        pm_out_folder = os.path.join(os.path.abspath(output_folder), "pipeline_manager")
        _LOGGER.info(f"Pipeline info folder = '{pm_out_folder}'")
        pm = pypiper.PipelineManager(
            name="bedboss-pipeline",
            outfolder=pm_out_folder,
            version=__version__,
            recover=True,
        )
        stop_pipeline = True
    else:
        stop_pipeline = False

    bbagent = BedBaseAgent(bedbase_config)

    validate_project(pep, BEDBOSS_PEP_SCHEMA_PATH)

    bedset_annotation = BedSetAnnotations(**pep.config).model_dump()
    skipper = Skipper(output_folder, pep.name)

    if rerun:
        skipper.reinitialize()

    if not lite:
        r_service = RServiceManager()
    else:
        r_service = None

    for i, pep_sample in enumerate(pep.samples):
        is_processed = skipper.is_processed(pep_sample.sample_name)
        if is_processed:
            m.print_success(
                f"Skipping {pep_sample.sample_name} : {is_processed}. Already processed."
            )
            processed_ids.append(is_processed)
            continue

        m.print_success(f"Processing sample {i + 1}/{len(pep.samples)}")
        _LOGGER.info(f"Running bedboss pipeline for {pep_sample.sample_name}")
        if pep_sample.get("file_type"):
            if pep_sample.get("file_type").lower() == "narrowpeak":
                is_narrow_peak = True
            else:
                is_narrow_peak = False
        else:
            is_narrow_peak = False
        try:

            if pep_sample.get("file_size", None):
                if int(pep_sample.file_size) > MAX_FILE_SIZE_QC:
                    raise QualityException(
                        f"File size {pep_sample.file_size} exceeds the maximum allowed size of {MAX_FILE_SIZE_QC}. Please provide a smaller file or set check_qc to False to skip this check."
                    )

            bed_id = run_all(
                input_file=pep_sample.input_file,
                input_type=pep_sample.input_type,
                genome=pep_sample.genome,
                name=pep_sample.sample_name,
                bedbase_config=bbagent,
                license_id=pep_sample.get("license_id") or license_id,
                narrowpeak=is_narrow_peak,
                chrom_sizes=pep_sample.get("chrom_sizes"),
                open_signal_matrix=pep_sample.get("open_signal_matrix"),
                other_metadata=pep_sample.to_dict(),
                outfolder=output_folder,
                rfg_config=rfg_config,
                check_qc=check_qc,
                ensdb=ensdb,
                just_db_commit=just_db_commit,
                force_overwrite=force_overwrite,
                update=update,
                upload_qdrant=upload_qdrant,
                upload_s3=upload_s3,
                upload_pephub=upload_pephub,
                universe=pep_sample.get("universe"),
                universe_method=pep_sample.get("universe_method"),
                universe_bedset=pep_sample.get("universe_bedset"),
                lite=lite,
                pm=pm,
                r_service=r_service,
            )

            processed_ids.append(bed_id)
            skipper.add_processed(pep_sample.sample_name, bed_id, success=True)

        except BedBossException as e:
            _LOGGER.error(f"Failed to process {pep_sample.sample_name}. See {e}")
            failed_samples.append(pep_sample.sample_name)
            skipper.add_failed(pep_sample.sample_name, f"{e}")

    if create_bedset:
        _LOGGER.info(f"Creating bedset from {pep.name}")
        run_bedbuncher(
            bedbase_config=bbagent,
            record_id=bedset_id or pep.name,
            bed_set=processed_ids,
            name=bedset_name or pep.name,
            output_folder=output_folder,
            description=pep.description,
            heavy=bedset_heavy,
            upload_pephub=upload_pephub,
            upload_s3=upload_s3,
            no_fail=no_fail,
            force_overwrite=force_overwrite,
            annotation=bedset_annotation,
            lite=lite,
        )
    else:
        _LOGGER.info(
            f"Skipping bedset creation. Create_bedset is set to {create_bedset}"
        )
    m.print_success(f"Processed samples: {processed_ids}")
    m.print_error(f"Failed samples: {failed_samples}")

    m.print_success(f"Processed samples: {processed_ids}")
    m.print_error(f"Failed samples: {failed_samples}")
    if stop_pipeline:
        pm.stop_pipeline()

    m.print_success(f"Processing of '{pep.name}' completed successfully")
    return None


@calculate_time
def reprocess_all(
    bedbase_config: Union[str, BedBaseAgent],
    output_folder: str,
    limit: int = 10,
    no_fail: bool = False,
    pm: pypiper.PipelineManager = None,
) -> None:
    """
    Run bedboss pipeline for all unprocessed beds in the bedbase

    Currently only beds with genomes hg19, hg38, and mm10 are processed.

    :param bedbase_config: bedbase configuration file path
    :param output_folder: output folder of the pipeline
    :param limit: limit of the number of beds to process
    :param no_fail: whether to raise an error if bedset was not added to the database
    :param pm: pypiper object

    :return: None
    """

    if not pm:
        pm_out_folder = os.path.join(os.path.abspath(output_folder), "pipeline_manager")
        _LOGGER.info(f"Pipeline info folder = '{pm_out_folder}'")
        pm = pypiper.PipelineManager(
            name="bedboss-pipeline",
            outfolder=pm_out_folder,
            version=__version__,
            recover=True,
        )
        stop_pipeline = True
    else:
        stop_pipeline = False

    r_service = RServiceManager()

    if isinstance(bedbase_config, str):
        bbagent = BedBaseAgent(config=bedbase_config)
    elif isinstance(bedbase_config, bbconf.BedBaseAgent):
        bbagent = bedbase_config
    else:
        raise BedBossException("Incorrect bedbase_config type. Exiting...")

    unprocessed_beds = bbagent.bed.get_unprocessed(
        limit=limit, genome=["hg38", "hg19", "mm10"]
    )

    bbclient = BBClient()
    failed_samples = []
    for i, bed_annot in enumerate(unprocessed_beds.results):
        bed_file = bbclient.load_bed(bed_annot.id)

        m.print_success(
            f"\n#### Processing sample: {i + 1} / {len(unprocessed_beds.results)} ####"
        )
        try:
            run_all(
                input_file=bed_file.path,
                input_type="bed",
                outfolder=output_folder,
                genome=bed_annot.genome_alias,
                bedbase_config=bbagent,
                name=bed_annot.name,
                license_id=bed_annot.license_id,
                rfg_config=None,
                check_qc=False,
                validate_reference=True,
                chrom_sizes=None,
                open_signal_matrix=None,
                ensdb=None,
                other_metadata=None,
                just_db_commit=False,
                update=True,
                upload_qdrant=True,
                upload_s3=True,
                upload_pephub=True,
                lite=False,
                universe=False,
                universe_method=None,
                universe_bedset=None,
                pm=pm,
                r_service=r_service,
            )
        except Exception as e:
            _LOGGER.error(f"Failed to process {bed_annot.name}. See {e}")
            if no_fail:
                raise BedBossException(f"Failed to process {bed_annot.name}. See {e}")

            failed_samples.append(
                {
                    "id": bed_annot.id,
                    "error": e,
                }
            )

    if failed_samples:
        date_now = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
        with open(
            os.path.join(output_folder, f"failed_samples_{date_now}.yaml"), "w"
        ) as file:
            yaml.dump(failed_samples, file)

            m.print_warning(f"Logs with failed samples are saved in {output_folder}")

    m.print_success(f"Processing completed successfully")

    print_values = dict(
        unprocessed_files=unprocessed_beds.count,
        processed_files=unprocessed_beds.limit,
        failed_files=len(failed_samples),
        success_files=unprocessed_beds.limit - len(failed_samples),
    )
    print_values["unprocessed_files_left"] = (
        print_values["unprocessed_files"] - print_values["processed_files"]
    )
    m.print_success(str(print_values))
    if stop_pipeline:
        pm.stop_pipeline()


@calculate_time
def reprocess_one(
    bedbase_config: Union[str, BedBaseAgent],
    output_folder: str,
    identifier: str,
    pm: pypiper.PipelineManager = None,
) -> None:
    """
    Run bedboss pipeline for one bed in the bedbase [Reprocess]

    :param bedbase_config: bedbase configuration file path
    :param output_folder: output folder of the pipeline
    :param identifier: bed identifier
    :param pm: pypiper object

    :return: None
    """

    if pm is None:
        pm_out_folder = os.path.join(os.path.abspath(output_folder), "pipeline_manager")
        _LOGGER.info(f"Pipeline info folder = '{pm_out_folder}'")
        pm = pypiper.PipelineManager(
            name="bedboss-pipeline",
            outfolder=pm_out_folder,
            version=__version__,
            recover=True,
        )
        stop_pipeline = True
    else:
        stop_pipeline = False

    if isinstance(bedbase_config, str):
        bbagent = BedBaseAgent(config=bedbase_config)
    elif isinstance(bedbase_config, bbconf.BedBaseAgent):
        bbagent = bedbase_config
    else:
        raise BedBossException("Incorrect bedbase_config type. Exiting...")

    bbclient = BBClient()

    bed_annot = bbagent.bed.get(identifier)
    bed_file = bbclient.load_bed(bed_annot.id)

    run_all(
        input_file=bed_file.path,
        input_type="bed",
        outfolder=output_folder,
        genome=bed_annot.genome_alias,
        bedbase_config=bbagent,
        name=bed_annot.name,
        license_id=bed_annot.license_id,
        rfg_config=None,
        check_qc=False,
        validate_reference=True,
        chrom_sizes=None,
        open_signal_matrix=None,
        ensdb=None,
        other_metadata=None,
        just_db_commit=False,
        update=True,
        upload_qdrant=True,
        upload_s3=True,
        upload_pephub=True,
        lite=False,
        universe=False,
        universe_method=None,
        universe_bedset=None,
        pm=None,
    )

    _LOGGER.info(f"Successfully processed {identifier}")

    if stop_pipeline:
        pm.stop_pipeline()


@calculate_time
def reprocess_bedset(
    bedbase_config: Union[str, BedBaseAgent],
    output_folder: str,
    identifier: str,
    no_fail: bool = True,
    heavy: bool = False,
):
    """
    Recalculate bedset from the bedbase

    :param bedbase_config: bedbase configuration file path
    :param output_folder: output folder of the pipeline
    :param identifier: bedset identifier
    :param no_fail: whether to raise an error if bedset was not added to the database
    :param heavy: whether to use heavy processing. Calculate plots for bedset

    :return: None
    """

    if isinstance(bedbase_config, str):
        bbagent = BedBaseAgent(config=bedbase_config)
    elif isinstance(bedbase_config, bbconf.BedBaseAgent):
        bbagent = bedbase_config
    else:
        raise BedBossException("Incorrect bedbase_config type. Exiting...")

    bedset_annot = bbagent.bedset.get(identifier)

    run_bedbuncher(
        bedbase_config=bbagent,
        record_id=bedset_annot.id,
        bed_set=bedset_annot.bed_ids,
        name=bedset_annot.name,
        output_folder=output_folder,
        description=bedset_annot.description,
        heavy=heavy,
        upload_pephub=False,
        upload_s3=heavy,
        no_fail=no_fail,
        force_overwrite=True,
        annotation={
            **bedset_annot.model_dump(
                exclude={
                    "bed_ids",
                }
            )
        },
        lite=False,
    )



import glob
import logging
import os
import time
import urllib.request
from functools import wraps
import gzip
from io import StringIO
import pandas as pd
from typing import Union
from gtars.models import RegionSet

import peppy
import requests
from bedms import AttrStandardizer
from pephubclient.files_manager import FilesManager
from peppy.const import SAMPLE_RAW_DICT_KEY
from pypiper import PipelineManager

from bedboss.refgenome_validator.main import ReferenceValidator
from bedboss.exceptions import QualityException
from bedboss.const import MIN_REGION_WIDTH, MAX_FILE_SIZE_QC

_LOGGER = logging.getLogger("bedboss")


def standardize_genome_name(
    input_genome: str,
    bedfile: Union[str, RegionSet] = None,
    reference_validator: ReferenceValidator = None,
) -> str:
    """
    Standardizing user provided genome

    :param input_genome: standardize user provided genome, so bedboss know what genome
    we should use
    :param bedfile: path to bed file
    :return: genome name string
    :param reference_validator: ReferenceValidator object, if None, a new one will be created
    :return: standardized genome name
    """

    if not reference_validator:
        reference_validator = ReferenceValidator()

    if bedfile:
        predicted_genome = reference_validator.predict(bedfile)[0]
        if predicted_genome:
            return predicted_genome

    if not isinstance(input_genome, str):
        input_genome = ""
    input_genome = input_genome.strip().lower()
    # TODO: we have to add more genome options and preprocessing of the string
    if input_genome == "hg38" or input_genome == "grch38":
        return "hg38"
    elif input_genome == "hg19" or input_genome == "grch37":
        return "hg19"
    elif input_genome == "mm10" or input_genome == "grcm38":
        return "mm10"
    elif input_genome == "mm9" or input_genome == "grcm37":
        return "mm9"
    else:
        return input_genome


def download_file(url: str, path: str, no_fail: bool = False) -> None:
    """
    Download file from the url to specific location

    :param url: URL of the file
    :param path: Local path with filename
    :param no_fail: If True, do not raise exception if download fails
    :return: NoReturn
    """
    _LOGGER.info(f"Downloading remote file: {url}")
    _LOGGER.info(f"Local path: {os.path.abspath(path)}")
    try:
        urllib.request.urlretrieve(url, path)
        _LOGGER.info("File downloaded successfully!")
    except Exception as e:
        _LOGGER.error("File download failed.")
        if not no_fail:
            raise e
        _LOGGER.error("File download failed. Continuing anyway...")


def get_genome_digest(genome: str) -> str:
    return requests.get(
        f"http://refgenomes.databio.org/genomes/genome_digest/{genome}"
    ).text.strip('""')


def example_bedbase_config():
    """
    Return example configuration for BedBase
    """
    return {
        "database": {
            "host": "localhost",
            "port": 5432,
            "user": "postgres",
            "password": "docker",
            "database": "bedbase",
            "dialect": "postgresql",
            "driver": "psycopg",
        },
        "qdrant": {
            "host": "localhost",
            "port": 6333,
            "api_key": None,
            "collection": "test_collection",
        },
        "server": {"host": "0.0.0.0", "port": 8000},
        "path": {
            "region2vec": "databio/r2v-encode-hg38",
            "vec2vec": "databio/v2v-geo-hg38",
            "text2vec": "sentence-transformers/all-MiniLM-L6-v2",
        },
        "access_methods": {
            "http": {
                "type": "https",
                "description": "HTTP compatible path",
                "prefix": "https://data2.bedbase.org/",
            },
            "s3": {
                "type": "s3",
                "description": "S3 compatible path",
                "prefix": "s3://data2.bedbase.org/",
            },
            "local": {
                "type": "https",
                "description": "How to serve local files.",
                "prefix": "/static/",
            },
        },
        "s3": {
            "endpoint_url": None,
            "aws_access_key_id": None,
            "aws_secret_access_key": None,
            "bucket": "bedbase",
        },
        "phc": {"namespace": "bedbase", "name": "bedbase", "tag": "latest"},
    }


def save_example_bedbase_config(path: str) -> None:
    """
    Save example configuration for BedBase

    :param path: path to the file
    """
    file_path = os.path.abspath(os.path.join(path, "bedbase_config.yaml"))
    FilesManager.save_yaml(example_bedbase_config(), file_path)
    _LOGGER.info(f"Example BedBase configuration saved to: {file_path}")


def standardize_pep(
    pep: peppy.Project, standard_columns: list = None, model: str = "BEDBASE"
) -> peppy.Project:
    """
    Standardize PEP file by using bedMS standardization model
    :param pep: peppy project
    :param standard_columns: list of columns to standardize

    :return: peppy project

    """
    if standard_columns is None:
        standard_columns = ["library_source", "assay", "genome", "species_name"]
    model = AttrStandardizer(model)
    suggestions = model.standardize(pep)

    changes = {}
    if suggestions is None:
        return pep
    for original, suggestion_dict in suggestions.items():
        for suggestion, value in suggestion_dict.items():
            if value > 0.9 and suggestion in standard_columns:
                if suggestion not in changes:
                    changes[suggestion] = {original: value}
                else:
                    if list(changes[suggestion].values())[0] < value:
                        changes[suggestion] = {original: value}

    raw_pep = pep.to_dict(extended=True)
    for suggestion, original_dict in changes.items():
        original_key = list(original_dict.keys())[0]
        if (
            suggestion not in raw_pep[SAMPLE_RAW_DICT_KEY]
            and original_key in raw_pep[SAMPLE_RAW_DICT_KEY]
        ):
            raw_pep[SAMPLE_RAW_DICT_KEY][suggestion] = raw_pep[SAMPLE_RAW_DICT_KEY][
                original_key
            ]
            del raw_pep[SAMPLE_RAW_DICT_KEY][original_key]

    return peppy.Project.from_dict(raw_pep)


def cleanup_pm_temp(pm: PipelineManager) -> None:
    """
    Cleanup temporary files from the PipelineManager

    :param pm: PipelineManager
    """
    if len(pm.cleanup_list_conditional) > 0:
        for cleandir in pm.cleanup_list_conditional:
            try:
                items_to_clean = glob.glob(cleandir)
                for clean_item in items_to_clean:
                    if os.path.isfile(clean_item):
                        os.remove(clean_item)
                    elif os.path.isdir(clean_item):
                        os.rmdir(clean_item)
            except Exception as e:
                _LOGGER.error(f"Error cleaning up: {e}")
        pm.cleanup_list_conditional = []


def calculate_time(func):
    @wraps(func)
    def wrapper(*args, **kwargs):

        # print(f"--> Arguments: {args}")
        # print(f"--> Keyword arguments: {kwargs}")

        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        execution_time = end_time - start_time

        hours, remainder = divmod(execution_time, 3600)
        minutes, seconds = divmod(remainder, 60)

        print(
            f"Function '{func.__name__}' executed in {int(hours)} hours, {int(minutes)} minutes, and {seconds:.2f} seconds"
        )

        return result

    return wrapper


def run_initial_qc(url: str, min_region_width: int = MIN_REGION_WIDTH) -> int:
    """
    Run initial QC on the bed file

    :param url: URL of the file
    :param min_region_width: Minimum region width threshold to pass the quality check. Default is 20

    :return: int. File size in bytes (0 if unable to determine)
    :raises: QualityException (includes file_size attribute)
    """
    _LOGGER.info(f"Running initial QC on the bed file: {url}")

    file_size = 0

    # Check file size before downloading content
    try:
        # Convert ftp:// to https:// for the HEAD request (e.g., NCBI FTP supports HTTPS)
        check_url = (
            url.replace("ftp://", "https://") if url.startswith("ftp://") else url
        )
        response = requests.head(check_url, allow_redirects=True)
        content_length = response.headers.get("Content-Length")
        if content_length:
            file_size = int(content_length)
        if file_size > MAX_FILE_SIZE_QC:
            file_size_mb = file_size / (1024 * 1024)
            max_size_mb = MAX_FILE_SIZE_QC / (1024 * 1024)
            raise QualityException(
                f"Initial QC failed for '{url}'. File size is '{file_size_mb:.2f} MB', where max file size is set to: '{max_size_mb:.0f} MB'",
                file_size=file_size,
            )
    except requests.RequestException as err:
        _LOGGER.warning(f"Unable to check file size: {err}. Continuing with QC...")

    try:
        with urllib.request.urlopen(url) as response:
            with gzip.GzipFile(fileobj=response) as f:
                content = f.read(10240).decode()  # Read first 10KB after decompression

        df = pd.read_csv(StringIO(content), sep="\t", header=None)
        mean_width = (df.iloc[:, 2] - df.iloc[:, 1])[:-1].mean()

    except Exception as err:
        _LOGGER.warning(
            "Unable to read the file, initial QC failed, but continuing anyway..."
            f"Error: {str(err)}"
        )
        return file_size

    if mean_width < min_region_width:
        raise QualityException(
            f"Initial QC failed for '{url}'. Mean region width is '{mean_width}', where min region width is set to: '{min_region_width}'",
            file_size=file_size,
        )

    _LOGGER.info(f"Initial QC passed for {url}")
    return file_size



#! /usr/bin/env python

import os
import sys

from setuptools import setup

PACKAGE_NAME = "bedboss"
REQDIR = "requirements"

# Additional keyword arguments for setup().
extra = {}

with open(f"{PACKAGE_NAME}/_version.py", "r") as versionfile:
    __version__ = versionfile.readline().split()[-1].strip("\"'\n")


def read_reqs(reqs_name):
    deps = []
    with open(os.path.join(REQDIR, f"requirements-{reqs_name}.txt"), "r") as file:
        for line in file:
            if not line.strip():
                continue
            deps.append(line)
    return deps


DEPENDENCIES = read_reqs("all")
extra["install_requires"] = DEPENDENCIES

scripts = None

with open("README.md") as f:
    long_description = f.read()

setup(
    name=PACKAGE_NAME,
    packages=[PACKAGE_NAME],
    version=__version__,
    description="Pipelines for genomic region file to produce bed files, "
    "and it's statistics",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    keywords="project, bioinformatics, sequencing, ngs, workflow",
    url="https://databio.org",
    authors=[
        "Oleksandr Khoroshevskyi",
        "Michal Stolarczyk",
        "Ognen Duzlevski",
        "Jose Verdezoto",
        "Bingjie Xue",
    ],
    author_email="khorosh@virginia.edu",
    license="BSD2",
    entry_points={
        "console_scripts": [
            "bedboss = bedboss.__main__:main",
        ],
    },
    package_data={PACKAGE_NAME: ["templates/*"]},
    include_package_data=True,
    test_suite="tests",
    tests_require=read_reqs("dev"),
    setup_requires=(
        ["pytest-runner"] if {"test", "pytest", "ptr"} & set(sys.argv) else []
    ),
    **extra,
)



import logging
import os
from pathlib import Path
from typing import Union

from refgenconf import CFG_ENV_VARS, CFG_FOLDER_KEY
from refgenconf import RefGenConf as RGC
from refgenconf import RefgenconfError, select_genome_config
from yacman.exceptions import UndefinedAliasError

from bedboss.const import DEFAULT_REFGENIE_PATH, REFGENIE_ENV_VAR

_LOGGER = logging.getLogger("bedboss")


def get_chrom_sizes(genome: str, rfg_config: Union[str, Path]) -> str:
    """
    Get chrom.sizes file with Refgenie.

    :param genome: genome name
    :return str: path to chrom.sizes file for the genome
    """

    _LOGGER.info("Determining path to chrom.sizes asset via Refgenie.")
    # initializing refginie config file
    rgc = get_rgc(rfg_config=rfg_config)

    try:
        # get local path to the chrom.sizes asset
        chrom_sizes = rgc.seek(
            genome_name=genome,
            asset_name="fasta",
            tag_name="default",
            seek_key="chrom_sizes",
        )
        _LOGGER.info(chrom_sizes)
    except (UndefinedAliasError, RefgenconfError):
        # if no local chrom.sizes file found, pull it first
        _LOGGER.info("Could not determine path to chrom.sizes asset, pulling")
        rgc.pull(genome=genome, asset="fasta", tag="default")
        chrom_sizes = rgc.seek(
            genome_name=genome,
            asset_name="fasta",
            tag_name="default",
            seek_key="chrom_sizes",
        )

    _LOGGER.info(f"Determined path to chrom.sizes asset: {chrom_sizes}")

    return chrom_sizes


def get_rgc(rfg_config: Union[str, Path] = None) -> RGC:
    """
    Get refgenie config file.

    :return str: rfg_config file path
    :return RGC: refgenie config object
    """
    if not rfg_config:
        _LOGGER.info("Creating refgenie genome config file...")
        cwd = os.getenv(REFGENIE_ENV_VAR, DEFAULT_REFGENIE_PATH)
        rfg_config = os.path.join(cwd, "genome_config.yaml")

    # get path to the genome config; from arg or env var if arg not provided
    refgenie_cfg_path = select_genome_config(filename=rfg_config, check_exist=False)

    if not refgenie_cfg_path:
        raise OSError(
            "Could not determine path to a refgenie genome configuration file. "
            f"Use --rfg-config argument or set the path with '{CFG_ENV_VARS}'."
        )
    if isinstance(refgenie_cfg_path, str) and not os.path.exists(refgenie_cfg_path):
        # file path not found, initialize a new config file
        _LOGGER.info(
            f"File '{refgenie_cfg_path}' does not exist. "
            "Initializing refgenie genome configuration file."
        )
        rgc = RGC(entries={CFG_FOLDER_KEY: os.path.dirname(refgenie_cfg_path)})
        rgc.initialize_config_file(filepath=refgenie_cfg_path)
    else:
        _LOGGER.info(
            "Reading refgenie genome configuration file from file: "
            f"{refgenie_cfg_path}"
        )
        rgc = RGC(filepath=refgenie_cfg_path)

    return rgc



import os

PKG_NAME = "bedboss"
DEFAULT_BEDBASE_API_URL = "https://api.bedbase.org"
# DEFAULT_BEDBASE_API_URL = "http://localhost:8000/api"

HOME_PATH = os.getenv("HOME")
if not HOME_PATH:
    HOME_PATH = os.path.expanduser("~")

OPEN_SIGNAL_FOLDER_NAME = "openSignalMatrix"
OPEN_SIGNAL_URL = "http://big.databio.org/open_chromatin_matrix/"

OS_HG38 = "openSignalMatrix_hg38_percentile99_01_quantNormalized_round4d.txt.gz"
OS_HG19 = "openSignalMatrix_hg19_percentile99_01_quantNormalized_round4d.txt.gz"
OS_MM10 = "openSignalMatrix_mm10_percentile99_01_quantNormalized_round4d.txt.gz"

BED_FOLDER_NAME = "bed_files"
BIGBED_FOLDER_NAME = "bigbed_files"
OUTPUT_FOLDER_NAME = "output"
BEDSTAT_OUTPUT = "bedstat_output"
QC_FOLDER_NAME = "bed_qc"

# bedqc
MAX_FILE_SIZE = 1024 * 1024 * 1024 * 2
MAX_FILE_SIZE_QC = 1024 * 1024 * 25  # 25 MB
MAX_REGION_NUMBER = 5000000
MIN_REGION_WIDTH = 10

# bedstat

# bedbuncher
DEFAULT_BEDBASE_CACHE_PATH = "./bedabse_cache"

BEDBOSS_PEP_SCHEMA_PATH = "https://schema.databio.org/pipelines/bedboss.yaml"
REFGENIE_ENV_VAR = "REFGENIE"

DEFAULT_REFGENIE_PATH = os.path.join(HOME_PATH, ".refgenie")

BED_PEP_REGISTRY = "databio/allbeds:bedbase"



cell_lines = {
    "1205lu": "1205Lu",
    "22rv1": "22Rv1",
    "293t": "293T",
    "2ftgh": "2fTGH",
    "3t3-l1": "3T3-L1",
    "46c mescs": "46C mESCs",
    "786-o": "786-O",
    "8988t": "8988T",
    "a172": "A172",
    "a375": "A375",
    "a498": "A498",
    "a549": "A549",
    "a673": "A673",
    "ag04450": "AG04450",
    "ag04503": "AG04503",
    "ag04504": "AG04504",
    "ag06561": "AG06561",
    "ag09309": "AG09309",
    "ag09319": "AG09319",
    "ag10803": "AG10803",
    "an3ca": "AN3CA",
    "aspc-1": "AsPC-1",
    "asc52telo": "ASC52telo",
    "be2c": "BE2C",
    "bin67": "BIN67",
    "bj": "BJ",
    "blaer1": "BLaER1",
    "bt474": "BT474",
    "bt549": "BT549",
    "c2c12": "C2C12",
    "c3h/10t1/2": "C3H/10T1/2",
    "c4-2b": "C4-2B",
    "caco-2": "Caco-2",
    "caki2": "Caki2",
    "calu-3": "Calu-3",
    "calu3": "Calu3",
    "colo320dm": "COLO320DM",
    "cov434": "COV434",
    "d341med": "D341Med",
    "dnd-41": "DND-41",
    "dohh2": "DOHH2",
    "du145": "DU145",
    "e14tg2a": "E14tg2a",
    "elf-1": "ELF-1",
    "f123 hybrid esc": "F123 hybrid ESC",
    "g401": "G401",
    "gbm002": "GBM002",
    "gm03348": "GM03348",
    "gm04504": "GM04504",
    "gm06990": "GM06990",
    "gm12865": "GM12865",
    "gm12878": "GM12878",
    "gm12891": "GM12891",
    "gm12892": "GM12892",
    "gm18505": "GM18505",
    "gm18870": "GM18870",
    "gm18907": "GM18907",
    "gm18909": "GM18909",
    "gm18951": "GM18951",
    "gm19025": "GM19025",
    "gm19099": "GM19099",
    "gm19239": "GM19239",
    "gm19324": "GM19324",
    "gm21526": "GM21526",
    "gm23248": "GM23248",
    "gm23338": "GM23338",
    # "h1": "H1",
    "h1-hesc": "H1-hESC",
    "h1975": "H1975",
    "h4": "H4",
    "h54": "H54",
    "h7": "H7",
    "h9": "H9",
    "hap-1": "HAP-1",
    "hap1": "HAP1",
    "hcec 1ct": "HCEC 1CT",
    "hcc1954": "HCC1954",
    "hcc3153": "HCC3153",
    "hct116": "HCT116",
    "hek293": "HEK293",
    "hek293t": "HEK293T",
    "hek293wt": "HEK293WT",
    "hela-s3": "HeLa-S3",
    "hela": "HeLa",
    "hepg2": "HepG2",
    "hff-myc": "HFF-Myc",
    "hg02798": "HG02798",
    "hg02970": "HG02970",
    "hg03045": "HG03045",
    "hl-60": "HL-60",
    "hoxb8": "HoxB8",
    "ht-29": "HT-29",
    "ht1080": "HT1080",
    "ht1376": "HT1376",
    "ht29": "HT29",
    "hudep-2": "HUDEP-2",
    "hudep2": "HUDEP2",
    "hues48": "HUES48",
    "hues6": "HUES6",
    "hues64": "HUES64",
    "huh-7": "HuH-7",
    "huh7": "Huh7",
    "huvec": "HUVEC",
    "imr-90": "IMR-90",
    "ips df 19.11": "iPS DF 19.11",
    "ips df 6.9": "iPS DF 6.9",
    "ips-20b": "iPS-20b",
    "ips-nihi7": "iPS-NIHi7",
    "ishikawa": "Ishikawa",
    "jurkat": "Jurkat",
    "k562": "K562",
    "karpas-422": "Karpas-422",
    "kc167": "Kc167",
    "kms-11": "KMS-11",
    "koptk1": "KOPTK1",
    "l1-s8": "L1-S8",
    "l1-s8r": "L1-S8R",
    "lncap": "LNCaP",
    "loucy": "Loucy",
    "luhmes": "LUHMES",
    "mcf 10a": "MCF 10A",
    "mcf-7": "MCF-7",
    "mcf7": "MCF-7",
    "mcf10a": "MCF10A",
    "mda-mb-231": "MDA-MB-231",
    "mda-mb-453": "MDA-MB-453",
    # "mg63": "MG63",
    "mkl-1": "MKL-1",
    "molm-13": "MOLM-13",
    "mm.1s": "MM.1S",
    "mv4-11": "MV4-11",
    "namalwa": "NAMALWA",
    "nalm6": "Nalm6",
    "nb4": "NB4",
    "nci-h1417": "NCI-H1417",
    "nci-h1436": "NCI-H1436",
    "nci-h187": "NCI-H187",
    "nci-h2081": "NCI-H2081",
    "nci-h2171": "NCI-H2171",
    "nci-h226": "NCI-H226",
    "nci-h446": "NCI-H446",
    "nci-h510a": "NCI-H510A",
    "nci-h524": "NCI-H524",
    "nci-h526": "NCI-H526",
    "nci-h82": "NCI-H82",
    "nci-h929": "NCI-H929",
    "nih-3t3": "NIH-3T3",
    "nt2/d1": "NT2/D1",
    "oci-ly1": "OCI-LY1",
    "oci-ly3": "OCI-LY3",
    "oci-ly7": "OCI-LY7",
    "panc1": "PANC1",
    "pc-3": "PC-3",
    "pc-9": "PC-9",
    "pfsk-1": "PFSK-1",
    "raw264.7": "RAW264.7",
    "rh30": "RH30",
    "riva": "RIVA",
    "rko": "RKO",
    "rpmi-7951": "RPMI-7951",
    "rpmi-8226": "RPMI-8226",
    "rpmi7951": "RPMI-7951",
    "rpmi8226": "RPMI-8226",
    "rues2": "RUES2",
    "rues02": "RUES02",
    "sh-sy5y": "SH-SY5Y",
    "shp-77": "SHP-77",
    "shp77": "SHP-77",
    "sjsa1": "SJSA1",
    "sk-mel-5": "SK-MEL-5",
    "sk-n-dz": "SK-N-DZ",
    "sk-n-mc": "SK-N-MC",
    "sk-n-sh": "SK-N-SH",
    "skbr3": "SKBR3",
    "sp-49": "SP-49",
    "su-dhl-6": "SU-DHL-6",
    "sum149": "SUM149",
    "sum159": "SUM159",
    "sw480": "SW480",
    "t47d": "T47D",
    "thp-1": "THP-1",
    "u2os": "U2OS",
    "u937": "U937",
    "uacc-257": "UACC-257",
    "ut-7": "UT-7",
    "vcap": "VCaP",
    "vero e6": "VERO E6",
    "weri-rb-1": "WERI-Rb-1",
    "wi38": "WI38",
    "wtc11": "WTC11",
    "zr-75-1": "ZR-75-1",
}


assay_map = {
    # ---- TF & Histone ChIP (most specific first) ----
    "tf chip-seq": "TF ChIP-seq",
    "histone chip-seq": "Histone ChIP-seq",
    # "h3k27ac": "H3K27ac",
    # "h3k27me3": "H3K27me3",
    # "h3k9me2": "H3K9me2",
    # "h3k9me3": "H3K9me3",
    # "h3k4me3": "H3K4me3",
    # "h3k4me2": "H3K4me2",
    # "h3k4me1": "H3K4me1",
    # "h3k36me3": "H3K36me3",
    # "h3k79me2": "H3K79me2",
    # "h3k79me3": "H3K79me3",
    # "h3k27me1": "H3K27me1",
    # "h3k9ac": "H3K9ac",
    # "h3k14ac": "H3K14ac",
    # "h3k18ac": "H3K18ac",
    # "h3k23ac": "H3K23ac",
    # "h4k20me1": "H4K20me1",
    # "h4k20me3": "H4K20me3",
    # "h2ak119ub": "H2AK119ub",
    # "h2bk120ub": "H2BK120ub",
    "biotin chip-seq": "Biotin ChIP-seq",
    # ---- Generic ChIP ----
    "chip-seq": "ChIP-seq",
    "chip seq": "ChIP-seq",
    # ---- CUT-based assays ----
    "cut&tag": "CUT&Tag",
    "cut&run": "CUT&RUN",
    # ---- ATAC variants ----
    "snail atac-seq": "SNAIL ATAC-seq",
    "csnail atac-seq": "cSNAIL ATAC-seq",
    "scatac-seq": "scATAC-seq",
    "roboatac": "roboATAC",
    "atac-seq": "ATAC-seq",
    "atac": "ATAC-seq",
    # ---- DNase ----
    "gm dnase-seq": "GM DNase-seq",
    "dnase-hypersensitivity": "DNase-Hypersensitivity",
    "dnase-seq": "DNase-seq",
    # ---- RNA & derivatives ----
    "rampage": "RAMPAGE",
    "cage": "CAGE",
    "rna-seq": "RNA-Seq",
    "ncrna-seq": "ncRNA-Seq",
    "mirna-seq": "miRNA-Seq",
    "rip-seq": "RIP-Seq",
    "eclip": "eCLIP",
    "icl ip-seq": "iCLIP-seq",
    # ---- Chromatin conformation ----
    "hi-c": "Hi-C",
    "chia-pet": "ChIA-PET",
    # ---- Methylation & accessibility ----
    "medip-seq": "MeDIP-Seq",
    "mbd-seq": "MBD-Seq",
    "bisulfite-seq": "Bisulfite-Seq",
    "faire-seq": "FAIRE-seq",
    "mnase-seq": "MNase-Seq",
    # ---- Specialized ----
    "end-seq": "END-seq",
    "tn-seq": "Tn-Seq",
    "bruuvseq": "bruUVseq",
    "bruseq": "BruSeq",
    "selex": "SELEX",
    "starr-seq": "STARR-seq",
    "microrna counts": "microRNA counts",
    "mitoperturb-seq": "MitoPerturb-Seq",
    # ---- Fallback ----
    "other": "OTHER",
}


antibody = [
    "H3K27ac",
    "H3K27me3",
    "H3K9me2",
    "H3K9me3",
    "H3K4me3",
    "H3K4me2",
    "H3K4me1",
    "H3K36me3",
    "H3K79me2",
    "H3K79me3",
    "H3K27me1",
    "H3K9ac",
    "H3K14ac",
    "H3K18ac",
    "H3K23ac",
    "H4K20me1",
    "H4K20me3",
    "H2AK119ub",
    "H2BK120ub",
]


def standardize_cell_line(cell_line: str) -> str:
    """
    Standardize cell line names to a consistent format.

    :param cell_line: The input cell line name.
    :return: The standardized cell line name.
    """

    key = cell_line.lower()
    if key not in cell_lines:
        return cell_line
    return cell_lines.get(key, cell_line)


def standardize_assay(assay: str) -> str:
    """
    Standardize assay names to a consistent format.

    :param assay: The input assay name.
    :return: The standardized assay name.
    """

    key = assay.lower()
    if key not in assay_map:
        return assay
    return assay_map.get(key, assay)


def find_assay(description: str) -> str:
    """
    Identify the assay type from a given description string.
    e.g. "encff285mhb_idr_thresholded_peaks_grch38_bed TF ChIP-seq from K562 (ENCSR861UWB)"
    will return "TF ChIP-seq"

    :param description: The input description string.
    :return: The identified assay type or "" (empty string) if not found.
    """
    description_lower = description.lower()
    for key in assay_map.keys():
        if key in description_lower:
            return assay_map[key]
    return ""


def find_cell_line(description: str) -> str:
    """
    Identify the cell line from a given description string.

    e.g. encff285mhb_idr_thresholded_k562_peaks_grch38_bed
    will return "K562"

    :param description: The input description string.
    :return: The identified cell line or "" (empty string) if not found.
    """
    description_lower = description.lower()
    for key in cell_lines.keys():
        if key in description_lower:
            return cell_lines[key]
    return ""



import logging
import os
import urllib.request

from bedboss.bbuploader.constants import DEFAULT_GEO_TAG, PKG_NAME

_LOGGER = logging.getLogger(PKG_NAME)


def build_gse_identifier(gse: str, geo_tag: str) -> str:
    """
    Build GSE identifier for database storage that includes the tag.
    For backward compatibility, 'samples' tag uses plain GSE name.

    :param gse: GEO series number (e.g., GSE123456)
    :param geo_tag: GEO tag ('samples' or 'series')
    :return: identifier for database (e.g., 'GSE123456' or 'GSE123456:series')
    """
    if geo_tag == DEFAULT_GEO_TAG:
        return gse
    return f"{gse}:{geo_tag}"


def download_file(file_url: str, local_file_path: str, force: bool = False) -> None:
    """
    Download file using ftp url

    :param file_url: downloading url
    :param local_file_path: path to the file or file name
    :param force: Rewrite if file exists
    :return: None
    """
    if force or not os.path.isfile(local_file_path):
        _LOGGER.info(f"Downloading file: '{file_url}' to: '{local_file_path}'")
        urllib.request.urlretrieve(file_url, local_file_path)
    else:
        _LOGGER.info(f"File {local_file_path} already exists. Skipping downloading.")


def create_gsm_sub_name(name: str) -> str:
    """
    Create gse subfolder name. e.g.
        gse123456 -> gsm123nnn
        gse123 -> gsennn
        gse1234-> gse1nnn
        gse1 -> gsennn

    ! This function was copied from geopephub utils

    :param name: gse name
    :return: gse subfolder name
    """

    len_name = len(name)

    if len_name <= 6:
        return """gsmnnn"""
    else:
        # return name[:6] + "n" * (len_name - 6)
        return name[:-3] + "n" * 3


def middle_underscored(s: str) -> str:
    """
    Return substring between the first and last underscore.
    If there are fewer than 3 underscores, return the original string unchanged.

    :param s: input string
    :return: substring between first and last underscore or original string
    """

    if s.count("_") < 3:
        return s
    parts = s.split("_")
    return "_".join(parts[1:-1])



# This module will serve to skip samples that were already processed.
import os
from typing import Union


class Skipper:
    def __init__(self, output_path: str, name: str):
        self.output_path = output_path
        self.name = name

        self.file_path = os.path.join(output_path, f"{name}.log")
        self.file_fail_log_path = os.path.join(output_path, f"{name}_fail.log")

        self.info = self._read_log_file(self.file_path)

    def is_processed(self, sample_name: str) -> Union[str, bool]:
        """
        Check if a sample was already processed.

        :param sample_name: name of the sample

        :return: digest of the sample
        """
        return self.info.get(sample_name, False)

    def add_processed(
        self, sample_name: str, digest: str, success: bool = False
    ) -> None:
        """
        Add a sample to the processed list.

        :param sample_name: name of the sample
        :param digest: digest of the sample
        :param success: if the processing was successful
        """
        # line = f"{sample_name}\t{digest}\t{'success' if success else 'failed'}\n"
        line = f"{sample_name},{digest}\n"
        with open(self.file_path, "a") as file:
            file.write(line)

        self.info[sample_name] = digest

    def _create_log_file(self, file_path: str) -> None:
        """
        Create a log file.

        :param file_path: path to the log file
        """
        with open(file_path, "w") as file:
            file.write("")

    def _read_log_file(self, file_path: str) -> dict:
        """
        Read the log file.

        :param file_path: path to the log file
        """
        data_dict = {}
        if os.path.exists(file_path):
            with open(file_path, "r") as file:
                for line in file:
                    # Split each line by whitespace
                    columns = line.strip().split(",")

                    # Assign the first column as the key, and the rest as the value
                    if len(columns) >= 2:  # Ensure there are at least three columns
                        key = columns[0]
                        value = columns[1]
                        data_dict[key] = value
        else:
            self._create_log_file(file_path)

        return data_dict

    def create_fail_log(self):
        """
        Create a log file for failed samples.
        """

        if not os.path.exists(self.file_fail_log_path):
            self._create_log_file(self.file_fail_log_path)

    def reinitialize(self):
        """
        Reinitialize the log file.
        """
        if os.path.exists(self.file_path):
            os.remove(self.file_path)
            self.info = self._read_log_file(self.file_path)
        else:
            self.info = self._read_log_file(self.file_path)

    def add_failed(self, sample_name: str, error: str = None):
        """
        Add a sample to the failed list.

        :param sample_name: name of the sample
        :param error: error message
        """

        line = f"{sample_name},{error}\n"
        with open(self.file_fail_log_path, "a") as file:
            file.write(line)
